"""Example plugins for filtering transactions.

These are various examples of how to filter entries in various creative ways.

IMPORTANT: These are not meant to tbe complete features, rather just experiments
in problem-solving using Beancount, work-in-progress that can be selectively
installed via a --plugin option, or one-offs to answer questions on the
mailing-list.
"""
__copyright__ = "Copyright (C) 2014, 2016  Martin Blais"
__license__ = "GNU GPLv2"
